"""
borgini._version
================

Contains the version of this package.

Allows for access to the version internally without cyclic imports
caused by accessing it through __init__.
"""
__version__ = "1.1.0"



"""
borgini.funcs
=============

Package entry point

Parse commandline arguments

Invoke classes, methods and functions
"""
from __future__ import annotations

import os
import re
import shutil
import subprocess
import sys
import typing as t

from . import config
from ._core import HOME, Data, PygmentPrint
from .parser import Catch


def get_configdir() -> str:
    """Get path to the config most suitable for active os and privilege.

    :return: Path to configuration dir for all profiles.
    """
    try:
        if os.getuid() == 0:
            return os.path.join("/etc", "xdg", "borgini")

        return os.path.join(HOME, ".config", "borgini")

    except AttributeError as err:
        raise AttributeError("Windows not currently supported") from err


CONFIGDIR = get_configdir()


def get_file_arg(
    namespace: t.Dict[str, str], files: t.Dict[str, str]
) -> str | None:
    """Detect that a file has been selected to edit or view.

    :param namespace: The ``ArgumentParser`` ``Namespace.__dict__``.
    :param files: The file's dictionary returned from ``data.Data``.
    :return: Return an absolute path or ``None``.
    """
    notfile = "editor", "dry", "select", "list", "remove"
    for key in namespace:
        if key not in notfile and namespace[key]:
            try:
                return files[key]

            except KeyError:
                return files[f"{key}.ini"]

    return None


def normalize_ntpath(path: str) -> str:
    """Format NT path to a Unix-like path.

    :param path: Path to format.
    :return: Returns a formatted path if running Windows otherwise the
        same path that came in.
    """
    normalized = "/".join(
        p[:-1] if re.match("^[a-zA-Z]:", p) else p
        for p in [p.lower() for p in path.split("\\")]
    )
    return normalized if path[0] == "/" else f"/{normalized}"


def edit_files(
    edit: str, file: str, pygments: PygmentPrint, dry: bool
) -> None:
    """Edit a config file with the editor of choice.

    :param edit: The editor to edit with.
    :param file: The path to the file to edit.
    :param pygments: Instantiated ``print.PygmentPrint`` class
        configured with user's style option.
    :param dry: Dry mode for when we do not want to execute the code.
    """
    file = normalize_ntpath(file)
    command = f"{edit} {file}"
    if dry:
        pygments.print(command, ini=False)
    else:
        subprocess.call(command, shell=True)


def read_file(filepath: str, pygments: PygmentPrint) -> None:
    """Read the file and print the output.

    If reading ``config.ini`` color the text with ini-style syntax,
    otherwise shell-like syntax.

    :param filepath: The file to read from.
    :param pygments: Instantiated ``print.PygmentPrint`` class.
        configured with user's style option.
    """
    ini = os.path.basename(filepath) == "config.ini"
    with open(filepath, encoding="utf-8") as file:
        lines = file.read()

    pygments.print(lines, ini=ini)


def edit_file(
    editor: str,
    namespace: t.Dict[str, str],
    files: t.Dict[str, str],
    pygments: PygmentPrint,
    dry: bool,
) -> None:
    """Call the editor to edit a file.

    If the argument passed does not correspond to an existing file print
    help. If an editor is not provided go on to simply print the file
    content.

    :param editor: The editor to edit the file with.
    :param namespace: ``argparse.ArgumentParser``'s
        ``Namespace.__dict__``.
    :param files: Dictionary of config file paths.
    :param pygments: Instantiated ``print.PygmentPrint`` class
        configured with user's style option.
    :param dry: Dry mode for when we do not want to execute the code.
    """
    filepath = get_file_arg(namespace, files)
    if editor:
        edit_files(editor, filepath, pygments, dry)  # type: ignore
        sys.exit(0)

    elif filepath:
        read_file(filepath, pygments)
        sys.exit(0)


def get_path(
    borguser: str, hostname: str, port: str, repopath: str, ssh: bool
) -> str:
    """Get the path to the backup repository location.

    If ``ssh`` is ``True`` this will return the configured ssh path on
    the remote system. If ``ssh`` is ``False`` the repository will need
    exist on the localhost.

    Note: The ``repopath`` setting in the ``DEFAULT`` section will
    be the directory the repository is in, not the actual repository
    i.e. ``/path/to/repopath/reponame`` not ``/path/to/repopath``.

    :param borguser: The remote's user invoking ``borg`` on that
        machine.
    :param hostname: The remote's hostname.
    :param port: The port that the remote allows ssh through.
    :param repopath: The path relative to root on the remote or the
        localhost.
    :param ssh: Boolean for whether using ssh or not.
    :return: A path configured based on the ``config.ini`` parameters.
    """
    return f"ssh://{borguser}@{hostname}:{port}{repopath}" if ssh else repopath


def set_passphrase(keyfile: str, catch: Catch) -> None:
    """Export the ``BORG_PASSPHRASE`` env var from a keyfile.

    :param keyfile: The absolute path to the passphrase keyfile.
    :param catch: ``parser.Catch`` - exit if necessary and inform
        user why.
    """
    if keyfile:
        if os.path.isfile(keyfile):
            with open(keyfile, encoding="utf-8") as file:
                passphrase = file.read().strip()

            os.environ["BORG_PASSPHRASE"] = passphrase
        else:
            catch.announce_keyfile()


def initialize_config(configpath: str, catch: Catch) -> config.Config:
    """If a config file does not exist create a default, else read.

    :param configpath: Path to config file.
    :param catch: Instantiated ``Catch`` object.
    :return: Instantiated ``ConfigParser`` object.
    """
    raw_config = config.RawConfig(configpath)
    if not os.path.isfile(configpath):
        raw_config.write_new_config()
        catch.announce_first_run()

    raw_config.read()
    return config.Config(raw_config)


def remove_profile(remove: t.List[str]) -> None:
    """Remove selected profile.

    :param remove: The profile entered to remove via
        ``argparse.ArgumentParser`` and the commandline.
    """
    profiles = ""
    if remove:
        plural = "s" if len(remove) > 1 else ""
        for profile in remove:
            profiledir = os.path.join(CONFIGDIR, profile)
            if os.path.isdir(profiledir):
                shutil.rmtree(profiledir)
                profiles += f"\n. {profile}"
            else:
                print(f"profile `{profile}' does not exist")
                sys.exit(1)

        print(f"removed profile{plural}:{profiles}")
        sys.exit(0)


def list_profiles(show_profiles: t.List[str], pygments: PygmentPrint) -> None:
    """If ``show_profiles`` then display a list of profiles that exist.

    :param show_profiles: Boolean switch from the commandline.
    :param pygments: Instantiated ``print.PygmentPrint`` class
        configured with user's style option.
    """
    if show_profiles:
        profiles = list(os.listdir(CONFIGDIR))
        profiles.insert(0, profiles.pop(profiles.index("default")))
        for profile in profiles:
            data = Data(CONFIGDIR, profile)
            configpath = data.get_path("config.ini")
            raw_config = config.RawConfig(configpath)
            try:
                raw_config.read()

            except NotADirectoryError:
                continue

            _config = config.Config(raw_config)
            obj = _config.dict["DEFAULT"]
            del obj["timestamp"]
            out = f"[{profile}]\n"
            for key, val in obj.items():
                out += f"{key} = {val}\n"

            pygments.print(out)
        sys.exit(0)


def initialize_datafiles(data: Data) -> None:
    """Write the default settings to the config files.

    :param data: ``data.Data``.
    """

    def comment(filetype):
        return (
            f"# --- {filetype} ---\n"
            "# This is an auto-generated list which should suite most users\n"
            f"# Remove any entries you do not want to {filetype} and add any "
            f"that you do\n"
            "#\n"
            "# . ensure you always use the absolute path for directories and "
            "files\n"
            "# . line and inline comments with `#' are supported\n"
            "#"
        )

    pathlists = (
        (
            f"{comment('include')}",
            "/home",
            "/root",
            "/var",
            "/usr/local",
            "/srv",
        ),
        (
            f"{comment('exclude')}",
            "/home/*/.cache/*",
            "/var/cache/*",
            "/var/tmp/*",
            "/var/run",
        ),
        (
            "# --- styles ---",
            "# Uncomment a single line to select style for syntax "
            "highlighting",
            "# `pygments' must be installed to use this feature",
            "#",
            "# . Install pygments by running:",
            "# . root:               `pip install pygments'",
            "# . user (recommended): `pip install pygments --user'",
            "#",
            '# STYLE="default"',
            '# STYLE="emacs"',
            '# STYLE="friendly"',
            '# STYLE="colorful"',
            '# STYLE="autumn"',
            '# STYLE="murphy"',
            '# STYLE="manni"',
            'STYLE="monokai"',
            '# STYLE="perldoc"',
            '# STYLE="pastie"',
            '# STYLE="borland"',
            '# STYLE="trac"',
            '# STYLE="native"',
            '# STYLE="fruity"',
            '# STYLE="bw"',
            '# STYLE="vim"',
            '# STYLE="vs"',
            '# STYLE="tango"',
            '# STYLE="rrt"',
            '# STYLE="xcode"',
            '# STYLE="igor"',
            '# STYLE="paraiso-light"',
            '# STYLE="paraiso-dark"',
            '# STYLE="lovelace"',
            '# STYLE="algol"',
            '# STYLE="algol_nu"',
            '# STYLE="arduino"',
            '# STYLE="rainbow_dash"',
            '# STYLE="abap"',
            '# STYLE="solarized-dark"',
            '# STYLE="solarized-light"',
            '# STYLE="sas"',
            '# STYLE="stata"',
            '# STYLE="stata-light"',
            '# STYLE="stata-dark"',
            '# STYLE="inkpot"',
        ),
    )
    data.initialize_data_files(pathlists)



"""
borgini.config
==============

All things ``config.ini``
"""
from __future__ import annotations

import configparser
import getpass
import os
import socket
import typing as t
from typing import Tuple


class RawConfig:
    """Contains the ``configparser.ConfigParser`` object.

    Write to and read from the ``config.ini`` file. The boolean values
    get written to the file as strings, and they get read from the file
    into the buffer as strings too. Not all values are as they should be
    when read into Python so this will get subclassed into
    ``config.Proxy`` first.

    :param configpath: The path to the config.ini file - this depends on
        the profile used and whether this is run as ``"$USER"`` or as
        root.
    """

    def __init__(self, configpath: str | os.PathLike) -> None:
        self.configpath = configpath
        self.parser = configparser.ConfigParser(interpolation=None)

    def _read_kwargs(self, **kwargs: t.Dict[str, t.Any]):
        self.parser.read_dict(kwargs)

    def _load_default_values(self) -> None:
        hostname = socket.gethostname()
        user = getpass.getuser()
        self._read_kwargs(
            DEFAULT={
                "reponame": hostname,
                "repopath": "None",
                "timestamp": "%Y-%m-%dT%H:%M:%S",
                "ssh": True,
                "prune": True,
            },
            SSH={"remoteuser": user, "remotehost": hostname, "port": "22"},
            BACKUP={
                "verbose": True,
                "stats": True,
                "list": True,
                "show-rc": True,
                "exclude-caches": True,
                "filter": "AME",
                "compression": "lz4",
            },
            PRUNE={
                "verbose": False,
                "stats": True,
                "list": True,
                "show-rc": True,
                "keep-daily": "7",
                "keep-weekly": "4",
                "keep-monthly": "6",
            },
            ENVIRONMENT={"keyfile": "None"},
        )

    def write_values(self) -> None:
        """Write values from ``ConfigParser`` to the config file."""
        with open(self.configpath, "w", encoding="utf-8") as configfile:
            self.parser.write(configfile)

    def write_new_config(self) -> None:
        """Load default values into the ``ConfigParser`` and write."""
        self._load_default_values()
        self.write_values()

    def read(self) -> None:
        """Read the ``config.ini`` file and avoid non-critical errors.

        Once the ``config.ini`` is read and loaded into the buffer write
        it back to the file as this class will filter out non-parsable
        configurations back to their default. If there is a key in the
        config.ini file that cannot be parsed skip reading it into
        buffer, as it will be removed once the config is subsequently
        written. Any new keys and configurations that may be added in
        the future will also be safely added to the config.
        """
        self._load_default_values()
        reader = configparser.ConfigParser(interpolation=None)
        reader.read(self.configpath)
        for section in reader:
            for key in dict(reader[section]):
                if section == "DEFAULT" or (
                    section != "DEFAULT" and key not in reader["DEFAULT"]
                ):
                    try:
                        self.parser[section][key] = reader[section][key]

                    except KeyError:
                        pass

        self.write_values()


class Proxy:
    """Subclass ``RawConfig`` to inherit the ``ConfigParser`` object.

    Translate the string values into boolean and ``NoneType`` values or
    some of the tests will not work i.e. ``None`` and ``False`` will be
    ``True`` as the strings ``"None"`` and ``"False"`` Inherit
    ``RawConfig`` after the ``parser.Catch`` string has identified
    run-time errors.

    :param raw_config: Instantiated ``RawConfig`` object containing the
        ``configparser.ConfigParser`` object as ``parser``.
    """

    def __init__(self, raw_config: RawConfig) -> None:
        self.parser = raw_config.parser
        self.sections = self.section_list()

    def section_list(self) -> t.List[str]:
        """Index the sections of the config.ini file.

        :return: List of config sections.
        """
        sections = self.parser.sections()
        sections.append("DEFAULT")
        return sections

    def _filter_default(self, section: str) -> t.Dict[str, t.Any]:
        if section == "DEFAULT":
            return dict(self.parser[section])

        return {
            k: v
            for k, v in self.parser[section].items()
            if k not in dict(self.parser["DEFAULT"])
        }

    def _raw_dict(self) -> t.Dict[str, t.Any]:
        return {s: self._filter_default(s) for s in self.sections}

    @staticmethod
    def _convert_null(obj: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        return {
            s: {k: v for k, v in obj[s].items() if v != "None"} for s in obj
        }

    def _get_boolean(self, section: str, key: str) -> bool:
        try:
            if key.startswith("keep-"):
                raise ValueError

            return self.parser.getboolean(section, key)

        except ValueError:
            return self._filter_default(section)[key]

    def _convert_booleans(self, obj: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        return {s: {k: self._get_boolean(s, k) for k in obj[s]} for s in obj}

    @staticmethod
    def _filter_null(raw_dict: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        return {k: v for k, v in raw_dict.items() if v}

    def convert_proxy(self) -> t.Dict[str, t.Any]:
        """Convert ``ConfigParser`` into python friendly dictionary.

        :return: Dictionary object.
        """
        raw_dict = self._raw_dict()
        truthy = self._convert_null(raw_dict)
        obj = self._convert_booleans(truthy)
        return self._filter_null(obj)


class Config(Proxy):
    """Final config object suitable for running with python methods.

    Subclass the ``config.Proxy`` class to inherit the translated
    config object from ``ConfigParser`` -> ``Dict``.

    :param raw_config: The ``configparser.ConfigParser`` object.
    """

    def __init__(self, raw_config: RawConfig) -> None:
        super().__init__(raw_config)
        self.dict = self.convert_proxy()

    def _get_section(self, section: str) -> t.Dict[str, t.Any]:
        return self.dict[section]

    def get_key(self, section: str, key: str) -> str | None:
        """Get a key from the dictionary object in ``self``.

        If the value is not found such as the ``BORG_PASSPHRASE``
        environment variable (as the string value ``"None"`` would have
        been omitted by ``config.Proxy``) return ``None`` in its
        place to avoid an expected error and carry on omitting the key.

        :param section: The primary key and the section from
            ``configparser.ConfigParser``.
        :param key: The key containing the configured value.
        :return: The value of the called key.
        """
        try:
            return self._get_section(section)[key]

        except KeyError:
            return None

    def get_keytuple(
        self, **kwargs: t.Tuple[str, ...]
    ) -> Tuple[str | bool | None, ...]:
        """Return multiple values at once from ``self.dict[section]``.

        Return as a tuple that can be unpacked by the keys passed to the
        method.

        :param kwargs: Sections to get.
        :return: A tuple of multiple any one or more values.
        """
        return tuple(
            self.get_key(s, k)
            for s in kwargs  # pylint: disable=consider-using-dict-items
            for k in kwargs[s]
        )

    @staticmethod
    def _get_flags(obj: t.Dict[str, str]) -> t.List[str]:
        return [f"--{k}" for k, v in obj.items() if v is True]

    @staticmethod
    def _get_keyvals(obj: t.Dict[str, str]) -> t.List[str]:
        return [f"--{k} {v}" for k, v in obj.items() if isinstance(v, str)]

    def return_all(self, section: str) -> t.Tuple[str, ...]:
        """Get all args belonging to a section.

        The (kw)args that are boolean flags only need to exist, as their
        existence in the script is the switch - so these are not
        returned as kwargs but rather args. Treat values that aren't
        boolean differently as the key and the value need to be
        included.

        :param section: The section from which the (kw)args should be
            retrieved.
        :return: A tuple of all switches prefixed with ``"--"`` and all
            kwargs.
        """
        obj = self.dict[section]
        args = self._get_flags(obj)
        args.extend(self._get_keyvals(obj))
        return tuple(args)



"""
borgini._main
=============

Contains package entry point.
"""
from __future__ import annotations

import shutil
import sys

from . import funcs, parser
from ._core import BorgBackup, Data, PygmentPrint

__version__ = "1.0.0"


class Parser(parser.RawParser):
    """Inherit ``RawParser`` which inherits ``argparse.ArgumentParser``.

    Catch any incompatible arguments. Instantiate and access the
    ``Namespace`` object from this class.
    """

    def __init__(self) -> None:
        super().__init__()
        self.editor = self.args.editor
        self.editor_opts = any(
            v for k, v in self.arg_groups["EDITOR"].__dict__.items()
        )
        self.message = self._get_message()
        self._exit_error()

    def _check_editor(self) -> str | None:
        # exit the process if the editor argument has been incorrectly
        # supplied
        if not shutil.which(self.editor):
            return parser.getcolor(
                f"EDITOR must be installed: `{self.editor}' cannot be found",
                code=1,
            )
        if not self.editor_opts:
            return parser.getcolor(
                "EDITOR must be followed by a file to edit", code=1
            )
        return None

    def _get_message(self) -> str | None:
        if self.editor:
            return self._check_editor()
        return None

    def _exit_error(self) -> None:
        if self.message:
            print(self.message, file=sys.stderr)
            self.print_help()
            sys.exit(1)


def main() -> None:
    """Main function for package."""
    _parser = Parser()
    args = _parser.args
    profile = args.select
    dry = args.dry
    data = Data(funcs.CONFIGDIR, profile)
    data.make_appdir()
    funcs.initialize_datafiles(data)
    pygments = PygmentPrint(data.files["styles"])
    catch = parser.Catch(profile)
    _configpath = data.get_path("config.ini")
    _config = funcs.initialize_config(_configpath, catch)
    funcs.remove_profile(args.remove)
    funcs.list_profiles(args.list, pygments)
    funcs.edit_file(args.editor, args.__dict__, data.files, pygments, dry)
    funcs.set_passphrase(
        _config.get_key("ENVIRONMENT", "keyfile"), catch  # type: ignore
    )
    catch.check_repopath_config(
        _config.get_key("DEFAULT", "repopath")  # type: ignore
    )
    fullpath = funcs.get_path(
        *_config.get_keytuple(  # type: ignore
            SSH=("remoteuser", "remotehost", "port"),
            DEFAULT=("repopath", "ssh"),
        )
    )
    borg = BorgBackup(
        fullpath,
        pygments,
        dry,
        *_config.get_keytuple(DEFAULT=("reponame", "timestamp")),
    )
    backup = _config.return_all("BACKUP")
    include = data.get_include()
    exclude = data.get_exclude()
    borg.create(backup, exclude, include)
    if _config.get_key("DEFAULT", "prune"):
        pruneargs = _config.return_all("PRUNE")
        borg.prune(pruneargs)



"""ini config for borg backup."""
from ._core import HOME, Data, PygmentPrint
from ._main import main
from ._version import __version__

__all__ = ["__version__", "HOME", "PygmentPrint", "Data", "funcs", "main"]



"""
borgini.parser
==============

Parse the commandline arguments and check arguments for errors and exit
informatively
"""
from __future__ import annotations

import argparse
import sys
import typing as t

from ._version import __version__


def getcolor(string: str, code: int) -> str:
    """Return a given string in color depending on the code provided.

    :param string: String to color.
    :param code: Corresponding ANSI escape code.
    :return: The colored string.
    """
    return f"\u001b[0;3{code};40m{string}\u001b[0;0m"


class RawParser(argparse.ArgumentParser):
    """Inherit ``ArgumentParser`` to be parsed from the commandline.

    Format a separate ``Namespace`` object to get arguments by group.
    """

    def __init__(self) -> None:
        # noinspection PyTypeChecker
        super().__init__(
            prog=getcolor("borgini", 6),
            formatter_class=lambda prog: argparse.HelpFormatter(
                prog, max_help_position=42
            ),
        )
        self._add_arguments()
        self._add_editor_args()
        self.args = self.parse_args()
        self.arg_groups: t.Dict[str | None, argparse.Namespace] = {}
        self._get_argument_group()
        self._version_request()

    def _add_editor_args(self) -> None:
        editor_args = self.add_argument_group(title="EDITOR")
        editor_args.add_argument(
            "-c", "--config", action="store_true", help="edit or view config"
        )
        editor_args.add_argument(
            "-i",
            "--include",
            action="store_true",
            help="edit or view list of files and directories to include",
        )
        editor_args.add_argument(
            "-e",
            "--exclude",
            action="store_true",
            help="edit or view list of files and directories to exclude",
        )
        editor_args.add_argument(
            "-s",
            "--styles",
            action="store_true",
            help="edit or view list of styles for syntax highlighting",
        )

    def _add_arguments(self) -> None:
        """Parse args from sys.argv."""
        self.add_argument(
            "editor",
            metavar="EDITOR",
            nargs="?",
            action="store",
            help="edit settings e.g. $ borgini vim --config",
        )
        self.add_argument(
            "-l", "--list", action="store_true", help="list existing profiles"
        )
        self.add_argument(
            "-d",
            "--dry",
            action="store_true",
            help="view the commandline arguments that would be run",
        )
        self.add_argument(
            "-v",
            "--version",
            action="store_true",
            help="show version and exit",
        )
        self.add_argument(
            "--select",
            action="store",
            default="default",
            metavar="PROFILE",
            help="create or select an alternative settings profile",
        )
        self.add_argument(
            "--remove",
            metavar="PROFILE",
            nargs="+",
            action="store",
            help="remove profile or profiles",
        )

    def _get_argument_group(self) -> None:
        for group in self._action_groups:
            # noinspection PyProtectedMember
            group_dict = {
                a.dest: getattr(self.args, a.dest, None)
                # pylint: disable=protected-access
                for a in group._group_actions
            }
            self.arg_groups[group.title] = argparse.Namespace(**group_dict)

    def _version_request(self) -> None:
        # print version if `--version` is passed to commandline
        if self.args.version:
            print(__version__)
            sys.exit(0)


class Catch:
    """Exit if it can be determined essential values are missing.

    Exit informatively.

    :param profile: Profile that the process is running under so that
        suggestion commands can be displayed.
    """

    def __init__(self, profile: str) -> None:
        self.profile = profile

    def show_command(self, arg: str) -> str:
        """Format and color the example commands provided.

        :param arg: The argument the command would need.
        :return: The formatted and colored example command.
        """
        return f". borgini EDITOR --{arg} --select {self.profile}"

    def check_repopath_config(self, repopath: str) -> None:
        """Exit process if the path to the backup repo not provided.

        Exit with a non-zero exit status.

        :param repopath: Path to back up repository or ``None``.
        """
        if not repopath:
            print(
                f"{getcolor('Path to repo not configured', 3)}\n"
                "Make all necessary changes to config before running this "
                "again\n"
                "You can do this by running the command:\n\n"
                f"{self.show_command('config')}",
                file=sys.stderr,
            )
            sys.exit(1)

    def announce_first_run(self) -> None:
        """Announce that a new default config file has been initialized.

        Not an error. Exit with zero exit status.
        """
        print(
            f"{getcolor('First run detected for profile', 3)}: "
            f"{getcolor(self.profile, 6)}\n"
            "Make all necessary changes to config before running this again\n"
            f"You can do this by running the command:\n\n"
            f"{self.show_command('config')}\n\n"
            "Default settings have been written to the `include' and "
            "`exclude' lists\n"
            "These can be edited by running:\n\n"
            f"{self.show_command('include')}\n"
            f"{self.show_command('exclude')}"
        )
        sys.exit(0)

    def announce_keyfile(self) -> None:
        """Announce keyfile has been created.

        If the keyfile section of the config file has been populated
        but the keyfile cannot be found let the user know as this would
        make it clearer as to why the backup might fail eventually if
        ``borgbackup`` prompts for a passphrase.

        If no keyfile is necessary then the field should be ``None``
        which this message would also notify the user about.
        """
        print(
            f'{getcolor("BORG_PASSPHRASE keyfile cannot be found", 3)}\n'
            f"attempting backup without keyfile\n\n"
            'add a valid keyfile or "None" to BORG_PASSPHRASE to stop '
            "receiving this message\n"
            "You can do this by running the command:\n",
            f"{self.show_command('config')}\n",
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['borgini']

package_data = \
{'': ['*']}

install_requires = \
['borgbackup>=1.1.15,<2.0.0', 'pygments>=2.6.1,<3.0.0']

entry_points = \
{'console_scripts': ['borgini = borgini.__main__:main']}

setup_kwargs = {
    'name': 'borgini',
    'version': '1.1.0',
    'description': 'ini config for borg backup',
    'long_description': "borgini\n=======\n.. image:: https://img.shields.io/badge/License-MIT-yellow.svg\n    :target: https://opensource.org/licenses/MIT\n    :alt: License\n.. image:: https://img.shields.io/pypi/v/borgini\n    :target: https://img.shields.io/pypi/v/borgini\n    :alt: pypi\n.. image:: https://github.com/jshwi/borgini/actions/workflows/ci.yml/badge.svg\n    :target: https://github.com/jshwi/borgini/actions/workflows/ci.yml\n    :alt: CI\n.. image:: https://github.com/jshwi/borgini/actions/workflows/codeql-analysis.yml/badge.svg\n    :target: https://github.com/jshwi/borgini/actions/workflows/codeql-analysis.yml\n    :alt: CodeQL\n.. image:: https://codecov.io/gh/jshwi/borgini/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jshwi/borgini\n    :alt: codecov.io\n.. image:: https://readthedocs.org/projects/borgini/badge/?version=latest\n    :target: https://borgini.readthedocs.io/en/latest/?badge=latest\n    :alt: readthedocs.org\n.. image:: https://img.shields.io/badge/python-3.8-blue.svg\n    :target: https://www.python.org/downloads/release/python-380\n    :alt: python3.8\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: black\n\nini config for borg backup\n--------------------------\n\nA wrapper to quickly get you started backing up with borg\n\nAn easy to use ini style and profile based format\n\nRequires:\n\n    - Python 3 >= 3.5.0, plus development headers\n    - OpenSSL >= 1.0.0, plus development headers\n    - libacl (which depends on libattr), both plus development headers\n    - liblz4 >= 1.7.0 (r129)\n    - libzstd >= 1.3.0\n    - libb2\n\n    For information on how to install these dependencies for Borg:\n    https://borgbackup.readthedocs.io/en/stable/installation.html\n\nInitialize the config\n\n.. code-block:: console\n\n    $ borgini\n    First run detected for profile: default\n    Make all necessary changes to config before running this again\n    You can do this by running the command:\n\n    . borgini EDITOR --config --select default\n\n    Default settings have been written to the ``include`` and ``exclude`` lists\n    These can be edited by running:\n\n    . borgini EDITOR --include --select default\n    . borgini EDITOR --exclude --select default\n..\n\n.. note::\n    the ``--select`` optional argument does not need to be passed if using the default profile\n\nEdit the config\n\n.. code-block:: console\n\n    $ borgini vim --config\n..\n\n.. note::\n    The selected editor is up to the user\n\n    The following would also work (provided they are installed)\n\n.. code-block:: console\n\n    $ borgini code --config\n    $ borgini gedit --config\n    $ borgini notepad --config\n..\n\nEnsure to make necessary changes to the ``DEFAULT`` section\n\nAnd ensure to configure the ``SSH`` section if an ssh repo is configured\n\nThe remaining configurations will suite most people\n\nIf you use the ``BORG_PASSPHRASE`` environment variable edit the ``ENVIRONMENT``\nsection to point to the keyfile\n\n.. note::\n    the file should contain one line and a password stored with safe read-write and ownership permissions\n\nEdit the include and exclude files\n\n.. code-block:: console\n\n    $ borgini vim --include  # add a list of paths to back up\n    $ borgini vim --exclude  # add a list of paths to exclude\n..\n\n.. note::\n    The exclude list can contain subdirectories and files listed within the include list\n\n    This will override their inclusion\n\nTo switch between profiles add ``--select PROFILE``\n\n.. code-block:: console\n\n    $ borgini vim --config  # edit default config\n    $ borgini vim --config --select profile2  # edit profile2's config\n    $ borgini vim --include --select profile2  # edit profile2's include file\n    $ borgini vim --exclude --select profile2  # edit profile2's exclude file\n    $ borgini --select profile2  # run profile2's backup\n..\n\nAdd the following for nightly backups at 12:00 to your crontab\n\n.. code-block:: console\n\n    $ 0 0 * * * /usr/local/bin/borgini\n    $ 0 0 * * * /usr/local/bin/borgini -s profile2  # easy for multiple repos\n..\n\n",
    'author': 'jshwi',
    'author_email': 'stephen@jshwisolutions.com',
    'maintainer': 'jshwi',
    'maintainer_email': 'stephen@jshwisolutions.com',
    'url': 'https://pypi.org/project/borgini/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
borgini._core
=============
"""
from __future__ import annotations

import datetime
import os
import pathlib
import shutil
import subprocess
import typing as t

from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter

# noinspection PyUnresolvedReferences
from pygments.lexers.configs import IniLexer

# noinspection PyUnresolvedReferences
from pygments.lexers.shell import BashLexer

HOME = str(pathlib.Path.home())


class BorgBackup:
    """``Borgbackup`` wrapper class.

    :param repopath: Path to the backup repository - local dir if
        ``ssh`` set to ``False``, remote server and repo dir if ``ssh``
        set to ``True``.
    :param pygments: Instantiated ``print.PygmentPrint`` class
        configured with user's style option.
    :param dry: if ``--dry`` is passed through the commandline then
        display the commandline arguments that would be passed to
        ``borgbackup`` without actually running it.
    :param args: Arguments parsed from the ``config.ini`` file with
        ``configparser.ConfigParser``.
    """

    def __init__(
        self, repopath: str, pygments: PygmentPrint, dry: bool, *args: t.Any
    ) -> None:
        self.repo = f"{repopath}/{args[0]}"
        self.pygments = pygments
        self.bin = shutil.which("borg")
        self.dry = dry if self.bin else True
        self.date = datetime.datetime.now().strftime(args[1])
        self.archive = f"{self.repo}::{args[0]}-{self.date}"

    @staticmethod
    def _separate_keep(
        args: t.List[str],
    ) -> t.Tuple[t.Tuple[t.Any, ...], t.Tuple[str, ...]]:
        keep = [
            args.pop(args.index(a))
            for a in list(args)
            if a.startswith("--keep-")
        ]
        return tuple(args), tuple(keep)

    def _run_borg(self, args: t.Tuple[str, ...]) -> None:
        borgargs = " ".join(args)
        command = f"{self.bin} {borgargs}"
        subprocess.call(command, shell=True)

    def _dry_mode(self, args: t.Tuple[str, ...]) -> None:
        borgargs = " ".join([f" {a}\n" for a in args])
        borg = self.bin if self.bin else "borg"
        self.pygments.print(f"\n{borg} {borgargs[1:-1]}", ini=False)

    def borg(self, *args: str) -> None:
        """Run ``borgbackup``.

        Run ``create`` to create a backup or ``prune`` to prune an old
        backup according to the ``--keep-*`` settings. If ``--dry`` has
        been passed through the commandline only.

        display the command that would occur - this won't run a backup
        or prune.

        :param args: Arguments for the borg command to receive, parsed
            from the ``config.ini``, the ``include`` and the ``exclude``
            files
        """
        if self.dry:
            self._dry_mode(args)
        else:
            self._run_borg(args)

    def create(
        self,
        args: t.Tuple[str, ...],
        exclude: t.Tuple[str, ...],
        include: t.Tuple[str, ...],
    ) -> None:
        """Run the ``borg create`` command.

        Includes several miscellaneous boolean arguments and key-values
        parsed from the ``config.ini`` file. Includes the ``exclude``
        arguments formatted as ``[--exclude ARG, ...]``. Includes the
        ``include`` arguments formatted as ``[ARG, ...]``.

        :param args: Flags that effect compression, whether stats are
            shown, the command is verbose etc...
        :param exclude: Directories or files to exclude parsed from the
            ``exclude`` list.
        :param include: Directories or files to include parsed from the
            ``include`` list - files and dirs will be overridden by
            exclude.
        """
        self.borg("create", *args, *exclude, f"{self.archive}", *include)

    def prune(self, args: t.Tuple[str, ...]) -> None:
        """Run the ``borg prune`` command.

        Includes several miscellaneous boolean arguments that can be
        further explored by running ``borgini --config``.

        :param args: Flags that effect compression, whether stats are
            shown, the command is verbose etc.
        """

        pruneargs, keep = self._separate_keep(list(args))
        self.borg("prune", *pruneargs, f"{self.repo}", *keep)


class Data:
    """Resolve the directory the configurations belong to.

    Get the list of files to include and exclude. Allow the files to
    contain comments and avoid parsing them.

    :param profile: The profile that ``borgini`` is being run under.
    """

    def __init__(self, appdir: str, profile: str) -> None:
        self.dirname = os.path.join(appdir, profile)
        self.files = self._get_file_obj()

    def _get_file_obj(self) -> t.Dict[str, str]:
        paths = "config.ini", "include", "exclude", "styles"
        return {p: os.path.join(self.dirname, p) for p in paths}

    def make_appdir(self) -> None:
        """Create the config directory for all the user's settings."""
        if not os.path.isdir(self.dirname):
            os.makedirs(self.dirname)

    def _get_paths(self, *args: str) -> t.Tuple[str, ...]:
        return tuple(self.files[arg] for arg in args)

    def initialize_data_files(
        self,
        pathlists: t.Tuple[
            t.Tuple[str, ...], t.Tuple[str, ...], t.Tuple[str, ...]
        ],
    ) -> None:
        """Initialize persistent data files.

        If the ``include`` or ``exclude`` files do not exist write the
            starter templates to file.

        :param pathlists: Tuple containing two tuples of lines to write
            to file.
        """
        for count, datafile in enumerate(
            self._get_paths("include", "exclude", "styles")
        ):
            if not os.path.isfile(datafile):
                with open(datafile, "w", encoding="utf-8") as file:
                    for path in pathlists[count]:
                        file.write(f"{path}\n")

    @staticmethod
    def _read_datafile(path: str) -> t.List[str]:
        with open(path, encoding="utf-8") as file:
            readfile = file.read().strip()

        return readfile.splitlines()

    def _parse_datafile(self, path: str) -> t.List[str]:
        files = self._read_datafile(path)

        # remove string starting at the hash symbol if inline
        # comment or remove entire line if the first index is the
        # hash symbol
        return [f"'{f.split('#')[0].strip()}'" for f in files if f[0] != "#"]

    def get_path(self, key: str) -> str:
        """Get the path of the file by calling its key.

        :param key: Name of the basename file
        :return: The file's absolute path
        """
        return self.files[key]

    def _get_files(self, path: str) -> t.List[str]:
        pathlist = []
        if os.path.isfile(path):
            pathlist.extend(self._parse_datafile(path))

        return pathlist

    def _format_include(self, path: str) -> t.Tuple[str, ...]:
        return tuple(self._get_files(path))

    def get_include(self) -> t.Tuple[str, ...]:
        """Directories and files to include from the ``include`` file.

        :return: Tuple of paths to include in backups for that profile
        """
        include_path = self.get_path("include")
        return self._format_include(include_path)

    @staticmethod
    def _format_exclude(pathlist: t.List[str]) -> t.Tuple[str, ...]:
        return tuple(f"--exclude {e}" for e in pathlist)

    def get_exclude(self) -> t.Tuple[str, ...]:
        """Return the values obtained from the ``exclude`` file.

        Unlike the ``include`` file this doesn't necessarily need to
        be populated

        Every item will automatically be prefixed with ``--exclude``
        for the ``borgbackup`` commandline

        :return: A lit of paths to exclude - this will override items in
            ``include``
        """
        exclude_path = self.get_path("exclude")
        exclude_list = self._get_files(exclude_path)
        return self._format_exclude(exclude_list)


class PygmentPrint:
    """Instantiate with the path to the config file for syntax styles.

    Class will read the file and maintain its config throughout the
    process.

    :param styles: The path to the ``styles`` config file.
    """

    def __init__(self, styles: str) -> None:
        self.styles = styles
        self.style = self.read_styles()

    def read_styles(self) -> str:
        """Read the ``styles`` file into the buffer.

        Parse the configuration that is uncommented and ignore the rest.

        :return: Style config.
        """
        with open(self.styles, encoding="utf-8") as file:
            contents = file.readlines()

        for content in contents:
            try:
                if content[0] != "#":
                    return content.split("=")[1].replace('"', "").strip()

            except IndexError:
                pass

        return "default"

    def print(self, string: str, ini: bool = True) -> None:
        """Print with ``pygments``.

        Read the string entered in method. Configure syntax highlighting
        for either shell or ini files and processes.

        :param string: What is to be printed.
        :param ini: ``True`` for printing ini files, ``False`` for
            shell.
        """
        if string:
            lexer = IniLexer if ini else BashLexer
            string = highlight(
                string, lexer(), Terminal256Formatter(style=self.style)
            )
            print(string)
